class CompareLSOutput:

    def __init__(self, a1, a2, distance_metric, result):
        """

        :param a1: Name of the first latent space
        :type a1: str
        :param a2: Name of the second latent space
        :type a2: str
        :param distance_metric: Name of the distance function
        :type distance_metric: str
        :param result: Distance
        :type result: str
        """
        self.a1 = a1
        self.a2 = a2
        self.distance_metric = distance_metric
        self.result = result

    def to_stdout(self):
        print(self.distance_metric + ' distance between ' + self.a1 + ' and ' + self.a2 + ': ' + self.result)

    def to_file(self, fname, ftype, mode):
        """

        :param fname: Output filename
        :type fname: str
        :param ftype: Output format
        :type ftype: str
        :param mode: Output mode
        :type mode: str
        """
        with open(fname, mode) as outf:
            if ftype == "text":
                outf.write(self.distance_metric + ' distance between ' + self.a1 + ' and ' + self.a2 + ': ' + self.result + '\n')
            else:
                outf.write(self.a1 + ',' + self.a2 + ',' + self.distance_metric + ',' + self.result + '\n')



import pandas as pd
import sys

if sys.version_info < (3, 9):
    # importlib.resources either doesn't exist or lacks the files()
    # function, so use the PyPI version:
    import importlib_resources
else:
    # importlib.resources has files(), so use that:
    import importlib.resources as importlib_resources


def print_families(a):
    print('Here is a list of protein families\' names:\n')
    pkg = importlib_resources.files("CLI")
    f = pkg / "seq_lengths.csv"
    family_list = pd.read_csv(f, usecols=['name']).squeeze("columns")

    print(*family_list, sep=', ')



import numpy as np

from . import aa_letters


def to_string(seqmat, remove_gaps=True):
    a = [''.join([aa_letters[np.argmax(aa)] for aa in seq]) for seq in seqmat]
    return [x.replace('-', '') for x in a] if remove_gaps else a

def greedy_decode_1d(arr1d):
    a = np.zeros(arr1d.shape)
    i = np.argmax(arr1d)
    a[i] = 1
    return a

def greedy_decode(pred_mat):
    return np.apply_along_axis(greedy_decode_1d, -1, pred_mat)

def _decode_nonar(generator, z, remove_gaps=False, alphabet_size=21, conditions=None):
    xp = generator.predict(z) if conditions is None else generator.predict([z, conditions])
    x = greedy_decode(xp)
    return to_string(x, remove_gaps=remove_gaps)

def _decode_ar(generator, z, remove_gaps=False, alphabet_size=21,
               sample_func=None, conditions=None):
    original_dim, alphabet_size = generator.output_shape[1], generator.output_shape[-1]
    x = np.zeros((z.shape[0], original_dim, alphabet_size))
    start = 0
    for i in range(start, original_dim):
        # iteration is over positions in sequence, which can't be parallelized
        pred = generator.predict([z, x]) if conditions is None else generator.predict([z, conditions, x])
        pos_pred = pred[:, i, :]
        if sample_func is None:
            pred_ind = pos_pred.argmax(-1) # convert probability to index
        else:
            pred_ind = sample_func(pos_pred)
        for j, p in enumerate(pred_ind):
            x[j, i, p] = 1

    seqs = to_string(x, remove_gaps=remove_gaps)
    return seqs

def batch_temp_sample(preds, temperature=1.0):
    batch_sampled_aas = []
    for s in preds:
        batch_sampled_aas.append(temp_sample_outputs(s, temperature=temperature))
    out = np.array(batch_sampled_aas)
    return out

def temp_sample_outputs(preds, temperature=1.0):
    # https://github.com/keras-team/keras/blob/master/examples/lstm_text_generation.py
    # helper function to sample an index from a probability array N.B. this works on single prob array (i.e. array for one sequence at one timestep)
    preds = np.asarray(preds).astype('float64')
    preds = np.log(preds) / temperature
    exp_preds = np.exp(preds)
    preds = exp_preds / np.sum(exp_preds)
    probas = np.random.multinomial(1, preds, 1)
    return np.argmax(probas)


class SearchLSOutput:

    def __init__(self, ls, distance_metric, closest, distance):
        """

        :param ls: Name of the latent space
        :type ls: str
        :param distance_metric: Name of the distance function
        :type distance_metric: str
        :param closest: Name of the closest protein family
        :type closest: str
        :param distance: Distance from the closest protein family
        :type distance: str
        """
        self.ls = ls
        self.distance_metric = distance_metric
        self.closest = closest
        self.distance = distance

    def to_stdout(self):
        print('The closest protein family to ' + self.ls + ' is ' + self.closest + ' with ' + self.distance_metric + ' distance: ' + self.distance)

    def to_file(self, fname, ftype, mode):
        """

        :param fname: Output filename
        :type fname: str
        :param ftype: Output format
        :type ftype: str
        :param mode: Output mode
        :type mode: str
        """
        with open(fname, mode) as outf:
            if ftype == "text":
                outf.write('The closest protein family to ' + self.ls + ' is ' + self.closest + ' with ' + self.distance_metric + ' distance: ' + self.distance + '\n')
            else:
                outf.write(self.ls + ',' + self.distance_metric + ',' + self.closest + ',' + self.distance + '\n')



from .load_files import load_family, load_ls_file, is_pf


class LSVectors:
    def __init__(self, pfnames):
        """

        :type pfnames: List[str]
        """
        self.names = pfnames
        self.ls_data = []
        self.ls_names = []
        self.load_vectors()

    def load_vectors(self):
        for ls in self.names:
            self.ls_names.append(ls)
            if is_pf(ls):
                self.ls_data.append(load_family(ls + '.txt'))
            else:
                self.ls_data.append(load_ls_file(ls))



#! /usr/bin/env python

# Author: Ramin Dehghanpoor

# import libraries
from .get_metric import get_distance_function
from .SearchLS import SearchLS
from .output_results import output_result


def ls_search(args):

    # get the arguments
    output_filename = args.output_file
    out_format = args.output_format
    out_mode = args.output_mode
    p_norm = args.p_norm  # The p-norm to apply for Minkowski - default is 2
    lat_spaces = args.latent_space
    
    # set the distance metric
    distance_function = get_distance_function(args.distance_metric)
    
    # find the closest latent space
    for lat_space in lat_spaces:
        res = SearchLS(lat_space, distance_function, p_norm).result
        output_result(res, output_filename, out_format, out_mode)



from scipy.spatial import distance
from .CompareLSOutput import CompareLSOutput


class CompareLS:
    def __init__(self, vectors, metric, p_norm):
        """

        :param vectors: Data from two latent spaces
        :type vectors: CLI.LSVectors.LSVectors
        :param metric: Distance function
        :type metric: function
        :param p_norm: The p-norm to apply for Minkowski
        :type p_norm: int
        """
        self.vectors = vectors
        self.metric = metric
        self.p_norm = p_norm
        self.result = self.do_compare()

    def do_compare(self):
        """

        :rtype: CompareLSOutput
        """
        # find distance between the vectors
        if self.metric == distance.minkowski:
            distance_result = self.metric(self.vectors.ls_data[0], self.vectors.ls_data[1], self.p_norm)
        else:
            distance_result = self.metric(self.vectors.ls_data[0], self.vectors.ls_data[1])

        # create CompareLSOutput object
        return CompareLSOutput(
            self.vectors.ls_names[0], self.vectors.ls_names[1], self.metric.__name__, str(distance_result))



#! /usr/bin/env python

# Author: Ramin Dehghanpoor

# import libraries
import argparse
from argparse_formatter import FlexiFormatter
from .search_seq import seq_search
from .family_list import print_families
from .search_lat import ls_search
from . import output_opt_parser, dist_opt_parser


def main():
    parser = argparse.ArgumentParser(description='''Find the closest protein family to a new latent space or protein sequence. 
    
Available metrics: 
    euclidean, minkowski, cityblock, sqeuclidean, cosine, correlation, hamming, jaccard, chebyshev, canberra, braycurtis, yule, dice, kulsinski, rogerstanimoto, russellrao, sokalmichener, sokalsneath

To see all the available protein families, run command:
    search names
            
Or you can find the closest protein family to first_new_latent_example.txt in cosine distance by running the command:
    search lat first_new_latent_example.txt -m cosine

Also you can find the closest family to a new protein sequence (for example new_sequence_example.txt) by running:
    search seq new_sequence_example.txt
    
    ''',
                                     formatter_class=FlexiFormatter)
    subparsers = parser.add_subparsers(dest='command')
    parser_ls = subparsers.add_parser('lat', help="Provide a new protein family latent space. The closest protein family to this new latent space will be shown.", parents=[output_opt_parser, dist_opt_parser])
    parser_ls.add_argument('latent_space', metavar="filename", help="The file name of a new latent space.", type=str, nargs='+')

    parser_seq = subparsers.add_parser('seq', help="Provide a protein sequence to get the closest protein family for this sequence.", parents=[output_opt_parser])
    parser_seq.add_argument('sequence', metavar="filename", help="The name of the file containing a protein sequence.", type=str, nargs='+')

    parser_names = subparsers.add_parser('names', help="Show available protein family names")

    parser_ls.set_defaults(func=ls_search)
    parser_seq.set_defaults(func=seq_search)
    parser_names.set_defaults(func=print_families)
    args = parser.parse_args()
    # print help text if nothing is selected
    if args.command is None:
        parser.print_help()
    else:
        args.func(args)

    # #parser.add_argument("-V",help="ndarray The variance vector for standardized Euclidean. Default: var(vstack([XA, XB]), axis=0, ddof=1)" ,dest="variance_vector", type=np.ndarray, default='None')
    # #parser.add_argument("-VI",help="ndarray The inverse of the covariance matrix for Mahalanobis. Default: inv(cov(vstack([XA, XB].T))).T" ,dest="inverse_covariance", type=np.ndarray, default='None')
    # args.help_text = parser.format_help()


if __name__ == "__main__":
    main()



# Author: Ramin Dehghanpoor

from scipy.spatial import distance


def get_distance_function(distance_metric):
    """

    :param distance_metric: Distance metric
    :type distance_metric: str
    :return: The distance function
    :rtype: function
    """
    # distance metric

    # set the distance metric
    if distance_metric == 'minkowski':
        distance_function = distance.minkowski
    elif distance_metric == 'cityblock':
        distance_function = distance.cityblock
    elif distance_metric == 'sqeuclidean':
        distance_function = distance.sqeuclidean
    elif distance_metric == 'cosine':
        distance_function = distance.cosine
    elif distance_metric == 'correlation':
        distance_function = distance.correlation
    elif distance_metric == 'hamming':
        distance_function = distance.hamming
    elif distance_metric == 'jaccard':
        distance_function = distance.jaccard
    elif distance_metric == 'chebyshev':
        distance_function = distance.chebyshev
    elif distance_metric == 'canberra':
        distance_function = distance.canberra
    elif distance_metric == 'braycurtis':
        distance_function = distance.braycurtis
    elif distance_metric == 'yule':
        distance_function = distance.yule
    elif distance_metric == 'dice':
        distance_function = distance.dice
    elif distance_metric == 'kulsinski':
        distance_function = distance.kulsinski
    elif distance_metric == 'rogerstanimoto':
        distance_function = distance.rogerstanimoto
    elif distance_metric == 'russellrao':
        distance_function = distance.russellrao
    elif distance_metric == 'sokalmichener':
        distance_function = distance.sokalmichener
    elif distance_metric == 'sokalsneath':
        distance_function = distance.sokalsneath
    else:
        distance_function = distance.euclidean
    return distance_function



import tensorflow as tf
from keras import backend as K


def aa_acc(prots_oh, reconstructed):
    x, fx = K.argmax(prots_oh, axis=-1), K.argmax(reconstructed, axis=-1)
    non_dash_mask = tf.greater(x, 0)
    aa_acc = K.sum(tf.cast(tf.boolean_mask(K.equal(x, fx), non_dash_mask), 'float32'))
    aa_acc /= K.sum(tf.cast(non_dash_mask, 'float32'))
    return aa_acc


from scipy.spatial import distance
from .SearchLSOutput import SearchLSOutput
from .load_files import load_ls_file, latent_space_list, load_family


class SearchLS:
    def __init__(self, ls, metric, p_norm):
        """

        :param ls: Name of a file containing latent space data
        :type ls: str
        :param metric: Distance function
        :type metric: function
        :param p_norm: The p-norm to apply for Minkowski
        :type p_norm: int
        """
        self.ls_name = ls
        self.ls_data = load_ls_file(ls)
        self.metric = metric
        self.p_norm = p_norm
        self.result = self.do_search()

    def do_search(self):
        """

        :rtype: CLI.SearchLSOutput.SearchLSOutput
        """
        closest_family = "none"
        min_dist = float("inf")
        for i in range(len(latent_space_list)):
            if self.metric == distance.minkowski:
                distance_result = self.metric(self.ls_data, load_family(latent_space_list[i]), self.p_norm)
            else:
                distance_result = self.metric(self.ls_data, load_family(latent_space_list[i]))

            if distance_result < min_dist:
                min_dist = distance_result
                closest_family = latent_space_list[i]
        return SearchLSOutput(self.ls_name, self.metric.__name__, closest_family[:-4], str(min_dist))



import numpy as np
import sys

if sys.version_info < (3, 9):
    # importlib.resources either doesn't exist or lacks the files()
    # function, so use the PyPI version:
    import importlib_resources
else:
    # importlib.resources has files(), so use that:
    import importlib.resources as importlib_resources

pkg = importlib_resources.files("CLI")
lspath = pkg / 'Latent_spaces'
s_length = pkg / "seq_lengths.csv"


# load data from protein family name
def load_family(fname):
    """

    :param fname: The name of a protein family
    :type fname: str
    :return: The data from its latent space file
    :rtype: numpy.ndarray
    """
    try:
        return np.loadtxt(lspath / fname)
    # if a file with that name exists but isn't valid, treat it as a new filename
    except ValueError:
        load_ls_file(fname)


# load data from new latent space file
def load_ls_file(fname):
    """

    :param fname: Name of a file containing latent space data
    :type fname: str
    :return: The latent space data from the file
    :rtype: numpy.ndarray
    """
    # try to read the file
    try:
        return np.loadtxt(fname)
    # if it isn't a valid latent space, give an error and quit
    except ValueError:
        print("Invalid file: " + fname)
        print("Files should contain 30 floats, each float in a separate line.")
        exit(2)
    # if the file doesn't exist or can't be read
    except IOError as err:
        exit(err)


def load_sequence(fname):
    try:
        seq_file = open(fname, "r+")
        return seq_file.read()
    # if the file doesn't exist or can't be read
    except IOError as err:
        exit(err)


def get_ls_list():
    """

    :return: List of protein family filenames
    :rtype: List[str]
    """
    ls_list = []
    for f in lspath.iterdir():
        ls_list.append(f.name)
    return ls_list


def is_pf(fname):
    """

    :param fname: Filename
    :type fname: str
    :return: If it's the name of a protein family
    :rtype: bool
    """
    if (lspath / (fname + '.txt')).is_file():
        return True
    return False


latent_space_list = get_ls_list()



import os
import urllib.request
from pathlib import Path
from zipfile import ZipFile
import numpy as np
import pandas as pd
from .utils import aa_letters
from .utils.data_loaders import to_one_hot
from .load_files import s_length, load_sequence
from keras.models import model_from_json
from .SearchSQOutput import SearchSQOutput
from tqdm import tqdm, trange
# import glob
# from keras import backend as K
import silence_tensorflow.auto


# Flag to disable progress bars
no_pbar = False

# get the sequence length for each protein family that we have
seq_lengths = pd.read_csv(s_length, usecols=['name', 'size'])
# list of all the trained networks. Each trained network belongs to a specific family
# networks_list = glob.glob('Trained_networks/*.h5')
aa_key = {l: i for i, l in enumerate(aa_letters)}


# get Amino Acid letter from the one hot encoded version of it
def get_AA(n):
    return list(aa_key.keys())[list(aa_key.values()).index(n)]


class DownloadProgressBar(tqdm):
    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)


def download_url(url, output_path):
    with DownloadProgressBar(unit='B', unit_scale=True, miniters=1, desc=url.split('/')[-1], disable=no_pbar) as t:
        urllib.request.urlretrieve(url, filename=output_path, reporthook=t.update_to)


def check_files():
    # if the trained network files don't already exist, download them
    if not Path('Trained_networks').exists():
        download_url('https://github.com/cfogel/Trained_networks/releases/download/Trained_networks/Trained_networks.zip', 'downloaded_file.zip')
        with ZipFile('downloaded_file.zip') as zf:
            for member in tqdm(zf.infolist(), desc='Extracting', leave=False, disable=no_pbar):
                zf.extract(member, 'Trained_networks')
        os.remove('downloaded_file.zip')


class SearchSQ:
    def __init__(self, seq):
        check_files()
        self.seq = seq
        self.result = self.do_search()

    def do_search(self):
        protein_seq = load_sequence(self.seq)
        max_acc = 0
        chosen_family = 'none'

        # loop over all the trained networks and find the one with highest reconstruction accuracy
        for i in trange(0, len(seq_lengths), total=len(seq_lengths), leave=False, disable=no_pbar):
            test_seq = protein_seq

            # skip the families with shorter protein sequence length
            if int(seq_lengths['size'][i]) < len(test_seq):
                continue

            # Not all families in sequence_lengths are in Trained_networks
            if Path('Trained_networks/' + seq_lengths['name'][i] + '.json').exists():

                # load the trained network
                json_file = open('Trained_networks/' + seq_lengths['name'][i] + '.json', 'r')
                loaded_model_json = json_file.read()
                json_file.close()
                loaded_model = model_from_json(loaded_model_json)

                # load weights into new model
                loaded_model.load_weights('Trained_networks/' + seq_lengths['name'][i] + '_weights.h5')

                # new_encoder = loaded_model.layers[1]
                # if int(loaded_model.layers[0].input_shape[1]/21) < len(test_seq):
                #    continue

                # add left and right gaps to get the same size sequence as the sequences used for training this network
                left_gaps = int((int(seq_lengths['size'][i]) - len(test_seq)) / 2)
                right_gaps = int(seq_lengths['size'][i]) - len(test_seq) - left_gaps
                test_seq = (left_gaps * '-') + test_seq + (right_gaps * '-')
                seq_length = len(test_seq)

                # use the one hot encoded version of the protein sequence
                single_msa_seq = [test_seq]
                x_test = to_one_hot(single_msa_seq)
                # y_test = K.argmax(x_test, axis=-1)
                x_test = x_test.reshape((len(x_test), np.prod(x_test.shape[1:])))

                # reconstruct the new protein sequence with the network
                a = loaded_model.predict(x_test, steps=1)
                a = a.reshape(len(single_msa_seq), seq_length, 21)

                # x is the reconstructed sequence
                x = np.vectorize(get_AA)(np.argmax(a, axis=-1))
                x = np.array(x).tolist()
                for k in range(0, len(x)):
                    x[k] = ''.join(x[k])
                count = 0

                # find the reconstruction accuracy
                for j in range(0, len(x[0])):
                    if x[0][j] == single_msa_seq[0][j]:
                        count = count + 1
                acc = count / len(x[0])

                # update the max accuracy and the chosen family name
                if acc > max_acc:
                    max_acc = acc
                    chosen_family = seq_lengths['name'][i]
        return SearchSQOutput(self.seq, chosen_family, str(max_acc))



import numpy as np
import pandas as pd
from . import aa_letters


def seq_to_one_hot(sequence, aa_key):
    arr = np.zeros((len(sequence), len(aa_key)))
    for j, c in enumerate(sequence):
        try:
            arr[j, aa_key[c]] = 1
        except KeyError as err:
            print("Invalid sequence letter")
            exit(err)
    return arr


def to_one_hot(seqlist, alphabet=aa_letters):
    aa_key = {l: i for i, l in enumerate(alphabet)}
    if type(seqlist) == str:
        return seq_to_one_hot(seqlist, aa_key)
    else:
        encoded_seqs = []
        for prot in seqlist:
            encoded_seqs.append(seq_to_one_hot(prot, aa_key))
        return np.stack(encoded_seqs)


def right_pad(seqlist, target_length=None):
    if target_length is None:
        return seqlist
    assert isinstance(target_length, int), 'Unknown format for argument padding'
    padded_seqlist = seqlist
    # handle padding either integer or character representations of sequences
    pad_char = '-' if isinstance(seqlist[0], str) else [0] if isinstance(seqlist[0], list) else None
    return [seq + pad_char * (target_length - len(seq)) for seq in seqlist]


def one_hot_generator(seqlist, conditions=None, batch_size=32,
                      padding=504, shuffle=True, alphabet=aa_letters):
    if type(seqlist) == pd.Series:
        seqlist = seqlist.values
    if type(seqlist) == list:
        seqlist = np.array(seqlist)
    if type(conditions) == list:
        conditions = np.array(conditions)

    n = len(seqlist)  # nb proteins
    prots_oh = None
    epoch = 0

    while True:
        # shuffle
        # print('Effective epoch {}'.format(epoch))
        if shuffle:
            perm = np.random.permutation(len(seqlist))
            prots = seqlist[perm]
            if conditions is not None:
                conds = conditions[perm]
        else:
            prots = seqlist
            conds = conditions

        for i in range(len(prots) // batch_size):
            batch = to_one_hot(right_pad(prots[i * batch_size:(i + 1) * batch_size], padding),
                               alphabet=alphabet)
            if conditions is not None:
                yield [batch, conds[i * batch_size:(i + 1) * batch_size]], batch
            else:
                yield batch, batch

        epoch += 1



import tensorflow as tf
from keras.layers import PReLU, ELU, LeakyReLU, Activation, Conv2DTranspose, \
     Conv1D as kConv1D, BatchNormalization, Add, Dropout, Reshape, Lambda
from keras import backend as K


BATCH_NORM = 'keras'

def _activation(activation, BN=True, name=None, momentum=0.9, training=None, config=BATCH_NORM):
    """
    A more general activation function, allowing to use just string (for prelu, leakyrelu and elu) and to add BN before applying the activation
    """

    def f(x):
        if BN and activation != 'selu':
            if config == 'keras':
                h = BatchNormalization(momentum=momentum)(x, training=training)
            elif config == 'tf' or config == 'tensorflow':
                h = BatchNorm(is_training=training)(x)
            else:
                raise ValueError('config should be either `keras`, `tf` or `tensorflow`')
        else:
            h = x
        if activation is None:
            return h
        if activation in ['prelu', 'leakyrelu', 'elu', 'selu']:
            if activation == 'prelu':
                return PReLU(name=name)(h)
            if activation == 'leakyrelu':
                return LeakyReLU(name=name)(h)
            if activation == 'elu':
                return ELU(name=name)(h)
            if activation == 'selu':
                return Selu()(h)
        else:
            h = Activation(activation, name=name)(h)
            return h

    return f

# BatchNorm
def BatchNorm(momentum=0.99, training=True):
    def batchnorm(x, momentum=momentum, training=training):
        return tf.layers.batch_normalization(x, momentum=momentum, training=training)

    def f(x):
        return Lambda(batchnorm, output_shape=tuple([xx for xx in x._keras_shape if xx is not None]))(x)

    return f

def Conv1D(filters, kernel_size, strides=1, padding='same', dilation_rate=1, activation=None, momentum=0.9,
           training=None, BN=True, config=BATCH_NORM,
           use_bias=False, kernel_initializer='glorot_uniform', bias_initializer='zeros', kernel_regularizer=None,
           bias_regularizer=None,
           activity_regularizer=None, kernel_constraint=None, bias_constraint=None, dropout=None, name=None, **kwargs):
    """BN after AtrousConvolution1D and BEFORE activation function"""

    def f(x):
        h = x
        if dropout is not None:
            h = Dropout(dropout)(h)
        h = kConv1D(filters,
                    kernel_size,
                    strides=strides,
                    padding=padding,
                    dilation_rate=dilation_rate,
                    activation=None,
                    use_bias=use_bias,
                    kernel_initializer=kernel_initializer,
                    bias_initializer=bias_initializer,
                    kernel_regularizer=kernel_regularizer,
                    bias_regularizer=bias_regularizer,
                    activity_regularizer=activity_regularizer,
                    kernel_constraint=kernel_constraint,
                    bias_constraint=bias_constraint,
                    **kwargs)(h)
        h = _activation(activation, BN=BN, name=name, momentum=momentum, training=training, config=config)(h)
        return h

    return f


def Deconv1D(filters, kernel_size, strides=2, padding='same', dilation_rate=1, activation="prelu", momentum=0.9,
             BN=True, config=BATCH_NORM,
             use_bias=False, training=None, kernel_initializer='glorot_uniform', bias_initializer='zeros',
             kernel_regularizer=None, bias_regularizer=None,
             activity_regularizer=None, kernel_constraint=None, bias_constraint=None, dropout=None, name=None):
    """`strides` is the upsampling factor"""

    def f(x):
        shape = list(x._keras_shape[1:])
        assert len(shape) == 2, "The input should have a width and a depth dimensions (plus the batch dimensions)"
        new_shape = shape[:-1] + [1] + [shape[-1]]
        h = Reshape(new_shape)(x)

        if dropout is not None:
            h = Dropout(dropout)(h)

        h = Conv2DTranspose(filters,
                            kernel_size,
                            strides=(strides, 1),
                            padding=padding,
                            dilation_rate=dilation_rate,
                            activation=None,
                            use_bias=use_bias,
                            kernel_initializer=kernel_initializer,
                            bias_initializer=bias_initializer,
                            kernel_regularizer=kernel_regularizer,
                            bias_regularizer=bias_regularizer,
                            activity_regularizer=activity_regularizer,
                            kernel_constraint=kernel_constraint,
                            bias_constraint=bias_constraint
                            )(h)
        h = _activation(activation, BN=BN, name=name, momentum=momentum, training=training, config=config)(h)
        shape = list(h._keras_shape[1:])
        new_shape = shape[:-2] + [filters]
        h = Reshape(new_shape)(h)
        return h

    return f


import sys
from .output_results import output_result


def seq_search(args):
    if sys.version_info[:3] == (3, 7, 6):
        # don't import function and dependencies if you don't need to
        from .SearchSQ import SearchSQ
        for ns in args.sequence:
            try:
                res = SearchSQ(ns).result
            except Exception as err:
                print("Error")
                exit(err)
            else:
                output_result(res, args.output_file, args.output_format, args.output_mode)

    else:
        print('sequence searches only available with python 3.7.6')



#! /usr/bin/env python

# Author: Ramin Dehghanpoor

# import libraries
import argparse
from argparse_formatter import FlexiFormatter
from .get_metric import get_distance_function
from .family_list import print_families
from .LSVectors import LSVectors
from .CompareLS import CompareLS
from . import output_opt_parser, dist_opt_parser
from .output_results import output_result


def run(args):
    # Get the arguments

    # Families
    family_list = args.family_name

    # Set the distance metric
    distance_function = get_distance_function(args.distance_metric)

    # The p-norm to apply for Minkowski
    p_norm = args.p_norm  # default is 2

    # Output arguments
    output_filename = args.output_file
    out_format = args.output_format
    out_mode = args.output_mode

    # when the user asks for the names of the proteins
    if family_list == ['list', 'names']:
        print_families(None)  # argument required in search but never used
        return

    else:
        # turn names into object containing data
        v = LSVectors(family_list)

    # find distance between the vectors, create CompareOutput object
    res = CompareLS(v, distance_function, p_norm).result
    output_result(res, output_filename, out_format, out_mode)


def main():
    parser = argparse.ArgumentParser(description='''Find the distance between fingerprints of two protein families. 
    
Available metrics: 
    euclidean, minkowski, cityblock, sqeuclidean, cosine, correlation, hamming, jaccard, chebyshev, canberra, braycurtis, yule, dice, kulsinski, rogerstanimoto, russellrao, sokalmichener, sokalsneath
       
As an example you can find the Euclidean distance between two families ATKA_ATKC and CDSA_RSEP by running the command:
    compare ATKA_ATKC CDSA_RSEP
    
    
Or if you want to find the cosine distance between two new latent spaces stored at first_new_latent_example.txt and second_new_latent_example.txt, you can run the command:
    compare first_new_latent_example.txt second_new_latent_example.txt -m cityblock
    
    
To show all available protein family names, run the command:
    compare list names
    
    ''',
                                     formatter_class=FlexiFormatter, parents=[output_opt_parser, dist_opt_parser])
    # parser.add_argument('--argument', default=None, help=''' ''')
    parser.add_argument("family_name", help="Protein family's name. Provide an existing protein family's name or the file name of a new latent space. Files should contain 30 floats, each float in a separate line.", metavar="protein_family", nargs=2, type=str)

    # parser.add_argument("-V",help="ndarray The variance vector for standardized Euclidean. Default: var(vstack([XA, XB]), axis=0, ddof=1)" ,dest="variance_vector", type=np.ndarray, default='None')
    # parser.add_argument("-VI",help="ndarray The inverse of the covariance matrix for Mahalanobis. Default: inv(cov(vstack([XA, XB].T))).T" ,dest="inverse_covariance", type=np.ndarray, default='None')
    parser.set_defaults(func=run)
    args = parser.parse_args()
    args.help_text = parser.format_help()
    args.func(args)


if __name__ == "__main__":
    main()



class SearchSQOutput:

    def __init__(self, seq, closest, accuracy):
        self.seq = seq
        self.closest = closest
        self.accuracy = accuracy

    def to_stdout(self):
        print('The closest protein family to ' + self.seq + ' is ' + self.closest + ' with average accuracy: ' + self.accuracy)

    # Not currently implemented in other modules
    def to_file(self, fname, ftype, mode):
        with open(fname, mode) as outf:
            if ftype == "text":
                outf.write('The closest protein family to ' + self.seq + ' is ' + self.closest + ' with average accuracy: ' + self.accuracy + '\n')
            else:
                outf.write(self.seq + ',' + self.closest + ',' + self.accuracy + '\n')



import argparse


output_opt_parser = argparse.ArgumentParser(add_help=False)
output_opt_parser.add_argument("-out", help="Output filename", dest="output_file", type=str, default="")
output_opt_parser.add_argument("-of", help="Output format. Default: text", dest="output_format", type=str, choices=["text", "csv"], default="text")
output_opt_parser.add_argument("-om", help="Output mode. Default: a", dest="output_mode", type=str, choices=['a', 'w'], default='a')

metrics = ['euclidean', 'minkowski', 'cityblock', 'sqeuclidean', 'cosine',
           'correlation', 'hamming', 'jaccard', 'chebyshev', 'canberra',
           'braycurtis', 'yule', 'dice', 'kulsinski', 'rogerstanimoto',
           'russellrao', 'sokalmichener', 'sokalsneath']

dist_opt_parser = argparse.ArgumentParser(add_help=False)
dist_opt_parser.add_argument("-m", help="Distance metric. Default: euclidean", metavar="DISTANCE_METRIC", dest="distance_metric", type=str, choices=metrics, default="euclidean")
dist_opt_parser.add_argument("-p", help="Scalar, The p-norm to apply for Minkowski, weighted and unweighted. Default: 2", dest="p_norm", type=int, default=2)



aa_letters = ['-', 'A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K',
              'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']

luxa_seq = 'MKFGNFLLTYQPPQFSQTEVMKRLVKLGRISEECGFDTVWLLEHHFTEFGLLGNPYVAAAYLLGATKKLNVGTAAIVLPTAHPVRQLEDVNLLDQMSK' \
           'GRFRFGICRGLYNKDFRVFGTDMNNSRALAECWYGLIKNGMTEGYMEADNEHIKFHKVKVNPAAYSRGGAPVYVVAESASTTEWAAQFGLPMILSWIIN' \
           'TNEKKAQLELYNEVAQEYGHDIHNIDHCLSYITSVDHDSIKAKEICRKFLGHWYDSYVNATTIFDDSDQTRGYDFNKGQWRDFVLKGHKDTNRRIDYSY' \
           'EINPVGTPQECIDIIQKDIDATGISNICCGFEANGTVDEIIASMKLFQSDVMPFLKEKQRSLLY'


import gzip


def load_gzdata(filename, one_hot=True):
    names, seqs = read_gzfasta(filename)
    if one_hot:
        seqs = to_one_hot(seqs)
    return names, seqs

def compress_file(filename):
    with open(filename, 'rb') as f:
        with open(filename+'.gz', 'wb') as gzout:
            b = gzip.compress(f.read())
            gzout.write(b)

def read_gzfasta(filepath, output_arr=False, encoding='utf-8'):
    names = []
    seqs = []
    seq = ''
    with gzip.open(filepath, 'rb') as fin:
        file_content = fin.read().decode(encoding)
        lines = file_content.split('\n')
        for line in lines:
            if line:
                if line[0] == '>':
                    if seq:
                        names.append(name)
                        if output_arr:
                            seqs.append(np.array(list(seq)))
                        else:
                            seqs.append(seq)
                    name = line[1:]
                    seq = ''
                else:
                    seq += line
            else:
                continue
        if seq: # handle last seq in file
            names.append(name)
            if output_arr:
                seqs.append(np.array(list(seq)))
            else:
                seqs.append(seq)
    if output_arr:
        seqs = np.array(seqs)
    return names, seqs

def read_fasta(filepath, output_arr=False):
    names = []
    seqs = []
    seq = ''
    with open(filepath, 'r') as fin:
        for line in fin:
            if line[0] == '>':
                if seq:
                    names.append(name)
                    if output_arr:
                        seqs.append(np.array(list(seq)))
                    else:
                        seqs.append(seq)
                name = line.rstrip('\n')[1:]
                seq = ''
            else:
                seq += line.rstrip('\n')
        if seq: # handle last seq in file
            names.append(name)
            if output_arr:
                seqs.append(np.array(list(seq)))
            else:
                seqs.append(seq)
    if output_arr:
        seqs = np.array(seqs)
    return names, seqs

def output_fasta(names, seqs, filepath):
    with open(filepath, 'w') as fout:
        for name, seq in zip(names, seqs):
            fout.write('>{}\n'.format(name))
            fout.write(seq+'\n')